==> 0.0.5.tar.gz: lib/src/field_list.dart <==
library fields.field_list;

import 'dart:async';
import 'dart:collection';
import 'dart:mirrors';
import 'field.dart';
import 'mirrors.dart';

// TODO(rms): https://code.google.com/p/dart/issues/detail?id=11186
typedef bool Predicate<E>(E element);

// TODO(rms): https://code.google.com/p/dart/issues/detail?id=11178
class FieldList extends ListBase<Field> {
  
  static final Comparator DEFAULT_COMPARATOR = 
      (Field a, Field b) => a.displayName.compareTo(b.displayName);
  
  static final Predicate DEFAULT_PREDICATE = (Field f) => true;
  
  final List<Field> _source = new List<Field>();
  final List<Field> _view = new List<Field>();
  
  final Field _parent;
  final _listeners = new List<StreamSubscription>();
  
  Comparator<Field> _comparator = DEFAULT_COMPARATOR;
  Comparator<Field> 
    get comparator => _comparator;
    set comparator(Comparator<Field> c) {
      assert(c != null);
      if (_comparator != c) {
        _comparator = c;
        for (Field f in _source) f.children.comparator = c;        
        refreshView();
      }
    }
    
  final _fieldChanged = new StreamController<Field>();
  Stream<Field> get fieldChanged => _fieldChanged.stream;
  
  Predicate<Field> _predicate = DEFAULT_PREDICATE;
  Predicate<Field>
    get predicate => _predicate;
    set predicate(Predicate<Field> p) {
      assert(p != null);
      if (_predicate != p) {
        _predicate = p;
        for (Field f in _source) f.children.predicate = p;        
        refreshView();
      }
    }
    
  int 
    get length => _view.length;
    set length(int length) => 
        throw new UnsupportedError('FieldList is not modifiable.');
      
  FieldList([this._parent]);
  
  Field operator [](int index) => _view[index];
  
  void operator []=(int index, Field value) => 
      throw new UnsupportedError('FieldList is not modifiable.');
  
  Map _parseMetadata(DeclarationMirror declaration) {
    assert(declaration != null);
    final metadata = new Map()
        ..[UseFieldEditor] = null
        ..[ReadOnly] = false
        ..[IgnoreField] = false
        ..[DisplayName] = null
        ..[FieldAnnotation] = <FieldAnnotation>[];
    declaration.metadata.forEach((InstanceMirror mirror) {
      final instance = mirror.reflectee;
      if (instance is UseFieldEditor) {
        // Disallow > 1 @UseFieldEditor
        assert(metadata[UseFieldEditor] == null);
        // TODO(rms): validate that the editor type isAssignableTo FieldEditor.
        metadata[UseFieldEditor] = reflectClass(instance.editorType);
      } else if (instance == ReadOnly) {
        metadata[ReadOnly] = true;
      } else if (instance == IgnoreField) {
        metadata[IgnoreField] = true;
      } else if (instance is DisplayName) {
        // Disallow > 1 @DisplayName
        assert(metadata[DisplayName] == null);
        metadata[DisplayName] = instance;
      } else if (instance is FieldAnnotation) {
        // Editor-specific annotations, store them in the field as-is.
        metadata[FieldAnnotation].add(instance);
      }
    });
    return metadata;
  }
  
  List<Field> _reflect(Object object) {
    final fields = new List<Field>();    
    final instanceMirror = reflect(object);
    // We reflect the ClassMirror from `Object.runtimeType` instead of getting
    // the ClassMirror from the `InstanceMirror.type` to support proxy objects.
    final classMirror = reflectClass(object.runtimeType);
    classMirror.members.forEach((key, member) {
      if (member is MethodMirror) { // method | getter | setter 
        if (member.isGetter 
            // TODO(rms): Symbol.validate throws illegal argument on _private
            // but we seem otherwise able to see the privates..is this a bug?
            && !member.isPrivate) {
          final setter = new Symbol('${MirrorSystem.getName(key)}='); 
          final metadata = _parseMetadata(member);
          if (!metadata[IgnoreField])
            fields.add(
                new Field(instanceMirror, 
                          key, 
                          typeOf(member.returnType), 
                          parent: _parent,
                          editor: metadata[UseFieldEditor],
                          annotations: metadata[FieldAnnotation],
                          displayName: metadata[DisplayName],
                          isPrivate: member.isPrivate,
                          isReadOnly: !classMirror.setters.containsKey(setter) 
                                      || metadata[ReadOnly]));
        }
      } else if (member is VariableMirror && !member.isStatic) {
        final metadata = _parseMetadata(member);
        if (!metadata[IgnoreField])
          fields.add(
              new Field(instanceMirror, 
                        key, 
                        typeOf(member.type), 
                        parent: _parent,
                        editor: metadata[UseFieldEditor],
                        annotations: metadata[FieldAnnotation],
                        displayName: metadata[DisplayName],
                        isPrivate: member.isPrivate,
                        isReadOnly: member.isFinal || metadata[ReadOnly]));
      }
    });
    return fields;
  }
  
  void refreshSource(Object object) {
    final fields = _reflect(object);
    fields.forEach((field) {
      // Here we consider fields equal if their name & type match.
      // We don't override == and hashCode for performance reasons; when we 
      // render a FieldListView we cache elements using Field keys, and in that 
      // situation Field identity works well and is the most efficient solution.
      final predicate = (f) => f.name == field.name && f.type == field.type;
      if (_source.any(predicate)) {
        _source.singleWhere(predicate)
            ..owner = field.owner
            ..refreshValue();
      }
      else {
        // Fields need to watch all of their siblings.
        fields
          .where((f) => f != field)
          .forEach((f) => f.values.listen((_) => field.refreshValue()));
        
        _source.add(field);
      }
    });
    refreshView();
  }
      
  void refreshView() {    
    _listeners.forEach((s) => s.cancel());
    _listeners.clear();
    
    _view
      ..clear()
      ..addAll(_source.where(predicate).toList()..sort(comparator)); 
    
    _view.forEach((Field f) {
      _listeners.add(f.values.listen((_) {
        if (_fieldChanged.hasListener) _fieldChanged.add(f);
      }));
      f.children.refreshView();
    });
  }
}

==> 0.0.8.tar.gz: lib/src/field_list.dart <==
library fields.field_list;

import 'dart:async';
import 'dart:collection';
import 'dart:mirrors';
import 'field.dart';

// TODO(rms): https://code.google.com/p/dart/issues/detail?id=11186
typedef bool Predicate<E>(E element);

// TODO(rms): https://code.google.com/p/dart/issues/detail?id=11178
class FieldList extends ListBase<Field> {
  
  static final Comparator DEFAULT_COMPARATOR = 
      (Field a, Field b) => a.displayName.compareTo(b.displayName);
  
  static final Predicate DEFAULT_PREDICATE = (Field f) => true;
  
  final List<Field> _source = new List<Field>();
  final List<Field> _view = new List<Field>();
  
  final Field _parent;
  final _listeners = new List<StreamSubscription>();
  
  Comparator<Field> _comparator = DEFAULT_COMPARATOR;
  Comparator<Field> 
    get comparator => _comparator;
    set comparator(Comparator<Field> c) {
      assert(c != null);
      if (_comparator != c) {
        _comparator = c;
        for (Field f in _source) f.children.comparator = c;        
        refreshView();
      }
    }
    
  final _fieldChanged = new StreamController<Field>();
  Stream<Field> get fieldChanged => _fieldChanged.stream;
  
  Predicate<Field> _predicate = DEFAULT_PREDICATE;
  Predicate<Field>
    get predicate => _predicate;
    set predicate(Predicate<Field> p) {
      assert(p != null);
      if (_predicate != p) {
        _predicate = p;
        for (Field f in _source) f.children.predicate = p;        
        refreshView();
      }
    }
    
  int 
    get length => _view.length;
    set length(int length) => 
        throw new UnsupportedError('FieldList is not modifiable.');
      
  FieldList([this._parent]);
  
  Field operator [](int index) => _view[index];
  
  void operator []=(int index, Field value) => 
      throw new UnsupportedError('FieldList is not modifiable.');
    
  List<Field> _reflect(Object object) {
    final fields = new List<Field>();    
    final instanceMirror = reflect(object);
    // We reflect the ClassMirror from `Object.runtimeType` instead of getting
    // the ClassMirror from the `InstanceMirror.type` to support proxy objects.
    final classMirror = reflectClass(object.runtimeType);
    classMirror.members.forEach((key, member) {
      if (member is MethodMirror) { // method | getter | setter 
        if (member.isGetter 
            // TODO(rms): Symbol.validate throws illegal argument on _private
            // but we seem otherwise able to see the privates..is this a bug?
            && !member.isPrivate) {
          final field = new Field.forGetter(
              instanceMirror, 
              key, 
              getter: member, 
              parent: _parent);
          if (field != null) {
            fields.add(field);
          }
        }
      } else if (member is VariableMirror && !member.isStatic) {
        final field = new Field.forVariable(
            instanceMirror, 
            key, 
            variable: member,
            parent: _parent);
        if (field != null) {
          fields.add(field);
        }
      }
    });
    return fields;
  }
  
  void refreshSource(Object object) {
    final fields = _reflect(object);
    fields.forEach((field) {
      // Here we consider fields equal if their name & type match.
      // We don't override == and hashCode for performance reasons; when we 
      // render a FieldListView we cache elements using Field keys, and in that 
      // situation Field identity works well and is the most efficient solution.
      final predicate = (f) => f.name == field.name && f.type == field.type;
      if (_source.any(predicate)) {
        _source.singleWhere(predicate)
            ..owner = field.owner
            ..refreshValue();
      }
      else {
        // Fields need to watch all of their siblings.
        fields
          .where((f) => f != field)
          .forEach((f) => f.values.listen((_) => field.refreshValue()));
        
        _source.add(field);
      }
    });
    refreshView();
  }
      
  void refreshView() {    
    _listeners.forEach((s) => s.cancel());
    _listeners.clear();
    
    _view
      ..clear()
      ..addAll(_source.where(predicate).toList()..sort(comparator)); 
    
    _view.forEach((Field f) {
      _listeners.add(f.values.listen((_) {
        if (_fieldChanged.hasListener) {
          _fieldChanged.add(f);
        }
      }));
      f.children.refreshView();
    });
  }
}
